'''
OOP Tree by Phil Thompson of Digital Maker CIC
https://digital-maker.co.uk/
for Raspberry Pi Pico running adafruit-circuitpython-raspberry_pi_pico-en_GB-7.3.0-beta.1
# implemented growing "nodes" 03/05/22
# implemented "polygon branches" getPolyPoints(node1, node2) 03/05/2022
# implemented "leaves" 04/05/2022
# implemented "jiggle" for random breeze on tree 04/05/2022
'''
import config
import displayio
import random
# vectorio is now part of CircuitPython (so no need for external library)
# but we still need to import it
from vectorio import Circle, Polygon
#from adafruit_display_shapes.line import Line
#from adafruit_display_text import label
import utilities

screen = config.screen
display = config.display

WIDTH, HEIGHT = config.WIDTH, config.HEIGHT

## make a random hex colour from "base rgb" + "variance rgb" values


def checkMax(x):
    # just make sure the value isn't over 255
    return max(0, min(x, 255))

def randRGB(base_rgb, rand_rgb):
    # natty function to join 2 rgb colours together (making sure values don't go outside (0 - 255))
    rr = checkMax(random.randint(base_rgb[0], base_rgb[0]+rand_rgb[0]))
    gg = checkMax(random.randint(base_rgb[1], base_rgb[1]+rand_rgb[1]))
    bb = checkMax(random.randint(base_rgb[2], base_rgb[2]+rand_rgb[2]))
    newCol = '0x{:02x}{:02x}{:02x}'.format(rr, gg, bb)
    # returns hex & gets converted to int(16) later
    return(newCol)

class Tree():
    ''' A Tree object creates nodes with random length, angle & coour variances'''
    
    wind = 100
    puff = 3
    
    def setWind(w):
        Tree.wind = w
        
    def setPuff(p):
        Tree.puff = p
        
    def getWind():
        return(Tree.wind, random.randrange(-Tree.puff, Tree.puff))

    treeBasics = {
        # handy pre-defined params for tree / plant types
        # 0 tree = Brown branches, green leaves
        0: {
            'x': int(WIDTH/2),
            'y': int(HEIGHT),
            'n': 6,
            'startLen': int(HEIGHT/3.5),
            'startRad': 6,
            'ang': 30,
            'plus': -90,
            'base_rgb': (240, 50, 50),
            'rand_rgb': (15, 100, 100),
            'leaf_base_rgb': (0, 150, 100),
            'leaf_rand_rgb': (15, 100, 50)
        }
    }
    
    def __init__(self,
                 fNode=None,
                 x=int(WIDTH/2),
                 y=HEIGHT,
                 n=7,
                 l=int(HEIGHT/1.5),
                 r=7,
                 ang=28,
                 plus=-90,
                 base_rgb=(15, 200, 100),
                 rand_rgb=(15, 55, 100)
                 ):
        # x,y position | n=nodes to generate | l=start length, r=radius of first node
        # ang=angle variance (+/-) | plus=angle to add to the nodes
        #  base rgb & variance rgb (for node colours)
        self.fNode = fNode  # follow node (for branches) should be a Node()
        self.x = x
        self.y = y
        self.numNodes = n
        self.startLen = self.l = l
        self.startRad = self.r = r
        self.ang = ang
        self.plus = plus
        self.base_rgb = base_rgb
        self.rand_rgb = rand_rgb
        self.scale = 0 # change to 0.01 for clean start (0.9 for testing)
        self.incs = 0.001
        self.maxScale = 1
        self.shrinker = 1.15
        # some arrays to hold nodes / lines / branches
        self.nodes = []
        self.lines = []
        self.branches = []
        # = set up an image group to draw everything into
        self.imGroup = displayio.Group()
        screen.append(self.imGroup)
        # use all these defaults to make the trunk
        self.setNodes()
        self.setLeaves()
        return

    def setNodes(self):
        # create a load of nodes from 0 to self.numNodes
        # newA (newAngle) holder
        # (build the "wobble" of a tree by making each node have a variable angle)
        newA = 0
        newR = self.startRad
        newL = self.startLen
        intA = int(self.ang)
        for node in range(self.numNodes):
            # create a random angle within constraints of
            newA = newA + random.randint(-intA, intA)
            newR = utilities.clamp(newR - (self.shrinker / 1.5), 1, HEIGHT / 4)
            newL = utilities.clamp(newL / (self.shrinker*1.2), 5, HEIGHT)
            # create a new Node & add it to self.nodes
            # passing self (Tree parent) & angle + rotator (plus)
            np = Node(self, ((newA * 1.1) + self.plus), newR, newL)
            self.nodes.append(np)
        # re-set the parents of nodes now we have created all the Nodes
        for i, node in enumerate(self.nodes):
            if i == 0:
                # first node, so "parent" = Tree
                if self.fNode == None:
                    node.setPar(self)
                else:
                    node.setPar(self.fNode)
            else:
                prevNode = self.nodes[i-1]
                nextLen = int(self.startLen - (i/self.shrinker))
                node.setPar(prevNode)

    def setLeaves(self):
        if self.fNode == None:
            # trunk can't have leaves until past node 1+
            startNode = 2 
        else:
            # to avoid overlapping branches on same nodes, start from 0+
            startNode = 0 

        for i, n in enumerate(self.nodes[startNode:]):
            n.makeLeaves(i)
        return
    
    # tree grow()
    def grow(self, dir=1):
        rrr, puff = Tree.getWind() # generate a random jiggle
        self.scale = utilities.clamp(
            self.scale + (self.incs*dir), 0.01, self.maxScale)
        # cycle through nodes (position & object)
        for i, node in enumerate(self.nodes):
            if self.fNode == None and i >= 1: # don't let a trunk node below 1 flex
                if random.randrange(rrr) <= 3:
                    node.setAngAdd(puff)
            node.grow()
            
        if self.branches:
            for branch in self.branches:
                branch.grow(dir)

    def newBranch(self, x):  # x = node(s) to follow
        if type(x) != list:
            x = [x]
        for nodeNum in x:
            if nodeNum > len(self.nodes): # trying to create a branch on a node that doesn't exist
                print("can't add a branch to node {} as it doesn't exist".format(nodeNum))
                x.remove(nodeNum)
        
        for i in x:
            # define who the new fNode (follow node) object is
            no = self.nodes[i]
            a = int(self.ang)
            branch = Tree(no,
                          int(no.x), int(no.y),
                          int(self.numNodes/1.5),
                          self.l/i,
                          no.r,
                          self.ang * 1.2,
                          self.plus + random.randrange(-a, a),
                          self.base_rgb,
                          self.rand_rgb
                          )
            self.branches.append(branch)

    def follow(self):
        self.x, self.y = self.fNode.x, self.fNode.y
    
    # Tree bloom (not used yet 04/05/2022)
    def bloom(self, bloomNodes=[]):
        if bloomNodes:
            for n in bloomNodes:
                self.nodes[n].bloom()
        else:
            self.nodes[-1].bloom()  # make the last node flower
            if self.branches:
                for b in self.branches:  # clycle through all the branches
                    b.nodes[-1].bloom()  # make last node on branch flower
    
    def getNodeLengthsUpToNode(self, toNode):
        # return the current total length of all the nodes l's * self.scale
        return(sum([obj.l for obj in self.nodes[0:self.nodes.index(toNode)]]))
        # return(sum([obj.l for obj in self.nodes[0:toNode]])*self.scale)
        
        
    def getInfo(self):
        for n in self.nodes:
            print(n.getInfo())

# external function to quickly generate 4 points for a polygon to connect 2 nodes


def getPolyPoints(n1, n2):  # make points based on 2 nodes
    # we need to make 2 points for node 1 & 2 points for node 2
    # we use node 1's angle & radius to determin the first 2 points
    n1r = int(n1.r)
    n2r = int(n2.r)
    if n1r < 3:
        n1r = 3
        n2r = 2
    p1 = utilities.calcXY(n1.ang - 90, int(n1r), (n1.x, n1.y))
    p2 = utilities.calcXY(n1.ang - 90, int(n2r), (n2.x, n2.y))
    # make the last 2 points based on node1's rotation & node2's radius * x,y
    p3 = utilities.calcXY(n1.ang + 90, int(n2r), (n2.x, n2.y))
    p4 = utilities.calcXY(n1.ang + 90, int(n1r), (n1.x, n1.y))

    return([p1, p2, p3, p4])

# node on a tree

class Node():
    # a node in a tree / branch (sub tree!) needing
    # parent, a=angle, r=radius, l=length (from previous node)
    def __init__(self, parent, a, r, l):
        self.tree = self.parent = parent
        self.x = self.parent.x
        self.y = self.parent.y
        self.ang = a  # direction of growth
        self.angFlex = 10
        self.angAdd = 0
        self.r = int(r)
        self.l = int(l)
        # colours for branches 
        self.pal = displayio.Palette(1)
        self.col = int(randRGB(self.parent.base_rgb, self.parent.rand_rgb), 16)
        self.pal[0] = self.col
        '''
        # commented out this section to see if we can avoid drawing nodes (speeds tree up)
        self.circ = Circle(pixel_shader=self.pal,
                           radius=self.r, x=self.x, y=self.y) # was oX, oY
        self.parent.imGroup.append(self.circ)
        '''
        # make an imgroup for this node for leaves
        self.leafImGroup = displayio.Group()
        screen.append(self.leafImGroup)
        
        self.flowers = self.leaves = []
        ''' # added this for different coloured lines to nodes
        # green lines created & passed back to Parent too
        #lC  line Colour / linePal = palette for line
        self.lC = int(randRGB((0, 200, 150), (30, 55, 55)), 16)
        self.linePal = displayio.Palette(1)
        self.linePal[0] = self.lC
        '''
        self.line = Polygon(pixel_shader=self.pal,
                            points=[(0, 0), (0, 0), (0, 0), (0, 0) ],
                            x=0, y=0)
        self.parent.imGroup.insert(0, self.line)
        self.parent.lines.append(self.line)
        return

    def setPar(self, par):
        self.parent = par
        if type(par) == Tree:
            self.l = 0  # root node - no length for growth
        self.grow()   
        
    def setAngAdd(self, a):
        # for a wee gust / movement in the tree
        newA = self.angAdd + a
        self.angAdd = utilities.clamp(newA, -self.angFlex, self.angFlex)
        
    #node leaves    
    def makeLeaves(self, i):
        # increase leaves per node the further from the start of the branch
        # clamp the amount of leaves per node between 2 & 8
        for l in range(utilities.clamp(i*2, 2, 9)): 
            self.leaves.append(Leaf(self))

    #node bloom (not used yet 04/05/22)
    def bloom(self):
        self.flowers.append(Flower(self, random.randint(3, 9)))    

    def getInfo(self):
        return(self.__dict__)

    #Node.grow()
    def grow(self):
        # try some random windy movement on this node!
        # calculate the new x,y from this node's angle & length from parent
        # and the parent's x,y location
        x, y = utilities.calcXY(
            # account for any "angle flex" (and clamp it to allowable range
            utilities.clamp(self.ang + self.angAdd, self.ang - self.angFlex, self.ang + self.angFlex),
            self.l * self.tree.scale,
            (self.parent.x, self.parent.y))
        # update the x,y in the Circle image too
        self.x = int(x)  # was self.x = self.circ.x = int(x)
        self.y = int(y) # was self.y = self.circ.y = int(y)
        # recalculate the "line" (polygon) from parent node to this node
        self.line.points = getPolyPoints(self.parent, self)
        # if this node has leaves, draw them ("follow") passing the scale (age)
        for l in self.leaves:
            l.follow()
            
class Leaf():
    '''
    Leaf Object created by node 
    Node has own "leafImGroup" to add leaf images to 
    (to avoid clogging up the tree imGroup (nodes & polygons))
    '''
    def __init__(self, parent):
        self.parent = parent
        self.a = random.randint(0,359) # random rotation from node
        self.l = random.randint(3, 15) # random distance from node
        self.rad = random.randint(2, 9) # potential leaf size
        self.pal = displayio.Palette(1)
        self.col = int(randRGB((20, 50, 50), (10, 150, 25)), 16)
        self.pal[0] = self.col
        self.leaf = Circle(
                        pixel_shader=self.pal,
                        radius=1, # start with 1px radius (can't have 0)
                        x=-100, y=int(HEIGHT*2)  # start the leaf babies offscreen
                    ) 
        self.parent.leafImGroup.append(self.leaf)
        return
        
    def follow(self):
        # check to see if the leaves should draw onstage based on collective length of nodes in tree / branch
        lenAwayFromRoot = self.parent.tree.getNodeLengthsUpToNode(self.parent)
        
        #if self.parent.y < int(HEIGHT - HEIGHT/4): # around 1/4 up from the bottom of the screen
        # see if calculating all the lens up to this point add up to more than 30!
        if lenAwayFromRoot*self.parent.tree.scale > lenAwayFromRoot/3:
            # pass the angle and the length of this leaf + parent x,y to get new x,y
            tempX, tempY = utilities.calcXY(                
                self.a, self.l,
                (self.parent.x, self.parent.y)
            )
            self.leaf.x, self.leaf.y = tempX, tempY
            self.leaf.radius = int(self.rad * self.parent.tree.scale) 
        else:
            self.leaf.x, self.leaf.y = -100, int(HEIGHT*2)
            self.leaf.radius = 1
        return        
           
# flower objects should be a leaf subclass / 04/05/2022     
class Flower():

    def __init__(self, parent, petals):
        self.parent = parent
        self.x, self.y = parent.x, parent.y
        self.petals = petals  # amount of petals allowed
        self.divs = 360.0/petals
        self.petalArray = []
        self.pal = displayio.Palette(1)
        self.col = int(randRGB((240, 240, 240), (15, 15, 15)), 16)
        self.pal[0] = self.col
        self.imGroup = displayio.Group()
        screen.append(self.imGroup)
        self.petSize = 2
        self.petLen = 7
        self.bloom()
        return

    def bloom(self):
        for p in range(self.petals):
            tempX, tempY = utilities.calcXY(
                # pass the angle (divs * position) and the length (self.petLen)
                self.divs*p, self.petLen,
                (self.parent.x, self.parent.y)
            )
            pet = Circle(pixel_shader=self.pal,
                         radius=self.petSize,
                         x=tempX, y=tempY)
            self.petalArray.append(pet)
            self.imGroup.append(pet)

    def follow(self):
        for i, p in enumerate(self.petalArray):
            p.x, p.y = utilities.calcXY(
                self.divs*i, self.petLen,
                (self.parent.x, self.parent.y)
            )
